

include \masm32\include\masm32.inc
includelib \masm32\lib\masm32.lib

NOT_ALLOWED_DOT equ 2
NOT_ALLOWED_TERMINAL equ 3
STACK_OVERFLOW equ 4
NOT_ENOUGH_OPERANDS equ 5
NOT_ENOUGH_OPERATORS equ 6
ZERO_DIV equ 7

PARAMSIZE typedef BYTE
FLOATSIZE typedef REAL10
MAX_NUM equ 256
TRUE   equ 1
FALSE  equ 0
PLUS equ '+'
MINUS equ '-'
MULF equ '*'
DIVF equ '/'
LEER equ ' '


.data?
floatStack FLOATSIZE(MAX_NUM) dup (?)


.code

;Param: input: String in UPNForm
;rckgabe: TRUE, wenn keine ungltigen Zeichen, ansonsten NOT_ALLOWED_TERMINAL 
;und in ESI Stringstelle
;Erlaubte Zeichen: 0-9 */+-. Leerzeichen
checkInput proc input:DWORD
	
	;check input auf unerlaubte Zeichen
  mov esi,input
  dec esi
  xor eax,eax
  CHECK_LOOP:
    add esi,1
    mov al,byte ptr [esi]
    test al,al
    jz CHECK_OUT
      cmp al,' '
      je CHECK_LOOP
      cmp al,'*'
        jb error
      cmp al,'/' 
        jg isdigit
      cmp al,','
        je error
        jmp CHECK_LOOP
      isdigit:
      cmp al,'9'
        jg error
      cmp al,'0'
        jb error
      
      jmp CHECK_LOOP
      error:
      mov eax,NOT_ALLOWED_TERMINAL
      ret
  CHECK_OUT:  
  mov eax,TRUE
  ret

checkInput endp

; Berechnet UPN Ausdruck
; Parameter: input: String in UPN Form, output: Pointer zu 8Byte Speicher,
; die Antwort wird in diesen Speicher geschrieben
; callback: optional, gibt in ST(0),ST(1) die Operanden und in EAX den Operator zurck
; so dass die Operation auch angezeigt werden kann
; Return: TRUE oder Fehlercode, in ESI Fehlerstelle (optionale Auswertung)
computeUPN proc input:DWORD, output:DWORD, callback:DWORD
LOCAL numend:DWORD
LOCAL num:REAL8
LOCAL dots:DWORD

  
  invoke checkInput,input
  cmp eax,TRUE
  je @f
    ret
  @@:  
  finit
  invoke stackNew,addr floatStack,MAX_NUM
  mov esi,input
  mov edi,esi ;edi soll numzeiger sein, esi wird dann das Ende des num-Teilstrings anzeigen
  dec esi
  
  
  xor eax,eax
  mov dots,0  
  PARSE_LOOP:   
    xor ebx,ebx
    add esi,1
    mov bl, byte ptr [esi]
    test bl,bl ;string zu ende?
      je PARSE_OUT    
    cmp bl,'.'
      jne @f  
      add dots,1
      cmp dots,1
      je PARSE_LOOP ;ansonsten mehr als 1 Punkt - unerlaubtes format
      mov eax,NOT_ALLOWED_DOT
      ret
    @@:  
    cmp bl,'9'
      jg ASCII2FLOAT
    cmp bl,'0'
      jb ASCII2FLOAT
    jmp PARSE_LOOP
    
    ASCII2FLOAT:
    mov dots,0
    ;in EDI ist Startadresse des Floatstrings, in ESI Endadresse
    ;push [esi] ;sichern
    cmp edi,esi ;0 Zeichen lange Zahl?
    je NO_FLOAT
    
    ;mov byte ptr [esi],0 ;mit 0 terminieren
    invoke StrToFloat, edi,addr num    
    cmp eax,esi
    jne NO_FLOAT
      invoke stackIsFull,addr floatStack
      cmp eax,TRUE
      jne @f
        mov eax,STACK_OVERFLOW
        ret
      @@:  
      fld num
      invoke floatStackPush,addr floatStack
    NO_FLOAT:
    
   
    mov edi,esi ;
    add edi,1 ;nchstes Zeichen vielleicht?
    
    cmp byte ptr [esi],LEER
    je PARSE_LOOP
    
   invoke floatStackPop,addr floatStack
   ;wenn Stack leer, so nicht genug operanden
   invoke stackIsEmpty,addr floatStack
   cmp eax,TRUE
   jne @f
      mov eax,NOT_ENOUGH_OPERANDS
      ret
   @@:  
  invoke floatStackPop,addr floatStack   
  
   mov bl, byte ptr [esi] ;CASE:   
   cmp callback,0
   je @f
     mov eax,ebx  ;Operation weitergeben
     call callback
   @@:
   cmp bl,PLUS
   jne @f
     fadd 
     jmp OP_OK
   @@:
   cmp bl,MINUS
   jne @f
     fsub ST(0),ST(1)
     jmp OP_OK
   @@:
   cmp bl,MULF
   jne @f
     fmul
     jmp OP_OK
   @@:
   cmp bl,DIVF  
   jne @f
     fdiv ST(0),ST(1)
     fstsw ax
     test ax,4
     jz OP_OK
       mov eax,ZERO_DIV
       ret
     
   @@:      
   OP_OK:  
   cmp callback,0
   je @f
     mov eax,'=' ;Ergebnis weitergeben
     call callback
   @@: 
   
    invoke floatStackPush, addr floatStack
   jmp PARSE_LOOP 
  PARSE_OUT:  
  
  invoke floatStackPop,addr floatStack
  invoke stackIsEmpty,addr floatStack
  cmp eax,TRUE
  je @f
      mov eax,NOT_ENOUGH_OPERATORS
      ret
  @@:  
  mov ecx,output
  fstp qword ptr[ecx]
  mov eax,TRUE
ret
computeUPN endp